# pythran export LTInversion(float64, float64, float64, float64, float64, float64)

import numpy as np


def LTInversion(t, lamda, alpha, beta, gama, log_epsilon):
    # Evaluation of the relevant poles
    theta = np.angle(lamda)
    kmin = np.ceil(-alpha/2. - theta/2./np.pi)
    kmax = np.floor(alpha/2. - theta/2./np.pi)
    k_vett = np.arange(kmin, kmax+1)
    s_star = np.abs(lamda)**(1./alpha) * \
        np.exp(1j*(theta+2*k_vett*np.pi)/alpha)

    # Evaluation of phi(s_star) for each pole
    phi_s_star = (np.real(s_star)+np.abs(s_star))/2

    # Sorting of the poles according to the value of phi(s_star)
    index_s_star = np.argsort(phi_s_star)
    phi_s_star = phi_s_star.take(index_s_star)
    s_star = s_star.take(index_s_star)

    # Deleting possible poles with phi_s_star=0
    index_save = phi_s_star > 1.0e-15
    s_star = s_star.repeat(index_save)
    phi_s_star = phi_s_star.repeat(index_save)

    # Inserting the origin in the set of the singularities
    s_star = np.hstack([[0], s_star])
    phi_s_star = np.hstack([[0], phi_s_star])
    J1 = len(s_star)
    J = J1 - 1

    # Strength of the singularities
    p = gama*np.ones((J1,), float)
    p[0] = max(0, -2*(alpha*gama-beta+1))
    q = gama*np.ones((J1,), float)
    q[-1] = np.inf
    phi_s_star = np.hstack([phi_s_star, [np.inf]])

    # Looking for the admissible regions with respect to round-off errors
    admissible_regions = \
        np.nonzero(np.bitwise_and(
            (phi_s_star[:-1] < (log_epsilon -
                                np.log(np.finfo(np.float64).eps))/t),
            (phi_s_star[:-1] < phi_s_star[1:])))[0]
    # Initializing vectors for optimal parameters
    JJ1 = admissible_regions[-1]
    mu_vett = np.ones((JJ1+1,), float)*np.inf
    N_vett = np.ones((JJ1+1,), float)*np.inf
    h_vett = np.ones((JJ1+1,), float)*np.inf

    # Evaluation of parameters for inversion of LT in each admissible region
    find_region = False
    while not find_region:
        for j1 in admissible_regions:
            if j1 < J1-1:
                muj, hj, Nj = OptimalParam_RB(
                    t, phi_s_star[j1], phi_s_star[j1+1], p[j1], q[j1], log_epsilon)
            else:
                muj, hj, Nj = OptimalParam_RU(
                    t, phi_s_star[j1], p[j1], log_epsilon)
            mu_vett[j1] = muj
            h_vett[j1] = hj
            N_vett[j1] = Nj
        if N_vett.min() > 200:
            log_epsilon = log_epsilon + np.log(10)
        else:
            find_region = True

    # Selection of the admissible region for integration which
    # involves the minimum number of nodes
    iN = np.argmin(N_vett)
    N = N_vett[iN]
    mu = mu_vett[iN]
    h = h_vett[iN]

    # Evaluation of the inverse Laplace transform
    k = np.arange(-N, N+1)
    u = h*k
    z = mu*(1j*u+1.)**2
    zd = -2.*mu*u + 2j*mu
    zexp = np.exp(z*t)
    F = z**(alpha*gama-beta)/(z**alpha - lamda)**gama*zd
    S = zexp*F
    Integral = h*np.sum(S)/2./np.pi/1j

    # Evaluation of residues
    ss_star = s_star[iN+1:]
    Residues = np.sum(1./alpha*(ss_star)**(1-beta)*np.exp(t*ss_star))

    # Evaluation of the ML function
    E = Integral + Residues
    if np.imag(lamda) == 0.:
        E = np.real(E)
    return E

# =========================================================================
# Finding optimal parameters in a right-bounded region
# =========================================================================


def OptimalParam_RB(t, phi_s_star_j, phi_s_star_j1, pj, qj, log_epsilon):
    # Definition of some constants
    log_eps = -36.043653389117154  # log(eps)
    fac = 1.01
    conservative_error_analysis = False

    # Maximum value of fbar as the ration between tolerance and round-off unit
    f_max = np.exp(log_epsilon - log_eps)

    # Evaluation of the starting values for sq_phi_star_j and sq_phi_star_j1
    sq_phi_star_j = np.sqrt(phi_s_star_j)
    threshold = 2.*np.sqrt((log_epsilon - log_eps)/t)
    sq_phi_star_j1 = min(np.sqrt(phi_s_star_j1), threshold - sq_phi_star_j)

    # Zero or negative values of pj and qj
    if pj < 1.0e-14 and qj < 1.0e-14:
        sq_phibar_star_j = sq_phi_star_j
        sq_phibar_star_j1 = sq_phi_star_j1
        adm_region = 1

    # Zero or negative values of just pj
    if pj < 1.0e-14 and qj >= 1.0e-14:
        sq_phibar_star_j = sq_phi_star_j
        if sq_phi_star_j > 0:
            f_min = fac*(sq_phi_star_j/(sq_phi_star_j1-sq_phi_star_j))**qj
        else:
            f_min = fac
        if f_min < f_max:
            f_bar = f_min + f_min/f_max*(f_max-f_min)
            fq = f_bar**(-1/qj)
            sq_phibar_star_j1 = (2*sq_phi_star_j1-fq*sq_phi_star_j)/(2+fq)
            adm_region = True
        else:
            adm_region = False

    # Zero or negative values of just qj
    if pj >= 1.0e-14 and qj < 1.0e-14:
        sq_phibar_star_j1 = sq_phi_star_j1
        f_min = fac*(sq_phi_star_j1/(sq_phi_star_j1-sq_phi_star_j))**pj
        if f_min < f_max:
            f_bar = f_min + f_min/f_max*(f_max-f_min)
            fp = f_bar**(-1./pj)
            sq_phibar_star_j = (2.*sq_phi_star_j+fp*sq_phi_star_j1)/(2-fp)
            adm_region = True
        else:
            adm_region = False

    # Positive values of both pj and qj
    if pj >= 1.0e-14 and qj >= 1.0e-14:
        f_min = fac*(sq_phi_star_j+sq_phi_star_j1) / \
            (sq_phi_star_j1-sq_phi_star_j)**max(pj, qj)
        if f_min < f_max:
            f_min = max(f_min, 1.5)
            f_bar = f_min + f_min/f_max*(f_max-f_min)
            fp = f_bar**(-1/pj)
            fq = f_bar**(-1/qj)
            if ~conservative_error_analysis:
                w = -phi_s_star_j1*t/log_epsilon
            else:
                w = -2.*phi_s_star_j1*t/(log_epsilon-phi_s_star_j1*t)
            den = 2+w - (1+w)*fp + fq
            sq_phibar_star_j = ((2+w+fq)*sq_phi_star_j + fp*sq_phi_star_j1)/den
            sq_phibar_star_j1 = (-(1.+w)*fq*sq_phi_star_j +
                                 (2.+w-(1.+w)*fp)*sq_phi_star_j1)/den
            adm_region = True
        else:
            adm_region = False

    if adm_region:
        log_epsilon = log_epsilon - np.log(f_bar)
        if not conservative_error_analysis:
            w = -sq_phibar_star_j1**2*t/log_epsilon
        else:
            w = -2.*sq_phibar_star_j1**2*t/(log_epsilon-sq_phibar_star_j1**2*t)
        muj = (((1.+w)*sq_phibar_star_j + sq_phibar_star_j1)/(2.+w))**2
        hj = -2.*np.pi/log_epsilon*(sq_phibar_star_j1-sq_phibar_star_j) \
            / ((1.+w)*sq_phibar_star_j + sq_phibar_star_j1)
        Nj = np.ceil(np.sqrt(1-log_epsilon/t/muj)/hj)
    else:
        muj = 0.
        hj = 0.
        Nj = np.inf

    return muj, hj, Nj


# =========================================================================
# Finding optimal parameters in a right-unbounded region
# =========================================================================
def OptimalParam_RU(t, phi_s_star_j, pj, log_epsilon):
    # Evaluation of the starting values for sq_phi_star_j
    sq_phi_s_star_j = np.sqrt(phi_s_star_j)
    if phi_s_star_j > 0:
        phibar_star_j = phi_s_star_j*1.01
    else:
        phibar_star_j = 0.01
    sq_phibar_star_j = np.sqrt(phibar_star_j)

    # Definition of some constants
    f_min = 1
    f_max = 10
    f_tar = 5

    # Iterative process to look for fbar in [f_min,f_max]
    while True:
        phi_t = phibar_star_j*t
        log_eps_phi_t = log_epsilon/phi_t
        Nj = np.ceil(phi_t/np.pi*(1. - 3*log_eps_phi_t /
                     2 + np.sqrt(1-2*log_eps_phi_t)))
        A = np.pi*Nj/phi_t
        sq_muj = sq_phibar_star_j*np.abs(4-A)/np.abs(7-np.sqrt(1+12*A))
        fbar = ((sq_phibar_star_j-sq_phi_s_star_j)/sq_muj)**(-pj)
        if (pj < 1.0e-14) or (f_min < fbar and fbar < f_max):
            break
        sq_phibar_star_j = f_tar**(-1./pj)*sq_muj + sq_phi_s_star_j
        phibar_star_j = sq_phibar_star_j**2
    muj = sq_muj**2
    hj = (-3*A - 2 + 2*np.sqrt(1+12*A))/(4-A)/Nj

    # Adjusting integration parameters to keep round-off errors under control
    log_eps = np.log(np.finfo(np.float64).eps)
    threshold = (log_epsilon - log_eps)/t
    if muj > threshold:
        if abs(pj) < 1.0e-14:
            Q = 0
        else:
            Q = f_tar**(-1/pj)*np.sqrt(muj)
        phibar_star_j = (Q + np.sqrt(phi_s_star_j))**2
        if phibar_star_j < threshold:
            w = np.sqrt(log_eps/(log_eps-log_epsilon))
            u = np.sqrt(-phibar_star_j*t/log_eps)
            muj = threshold
            Nj = np.ceil(w*log_epsilon/2/np.pi/(u*w-1))
            hj = np.sqrt(log_eps/(log_eps - log_epsilon))/Nj
        else:
            Nj = np.inf
            hj = 0

    return muj, hj, Nj



import numpy as np
from scipy.special import gamma
from .ml_internal import LTInversion


def ml(z, alpha, beta=1., gama=1.):
    eps = np.finfo(np.float64).eps
    if np.real(alpha) <= 0 or np.real(gama) <= 0 or np.imag(alpha) != 0. \
       or np.imag(beta) != 0. or np.imag(gama) != 0.:
        raise ValueError(
            'ALPHA and GAMA must be real and positive. BETA must be real.')
    if np.abs(gama-1.) > eps:
        if alpha > 1.:
            raise ValueError('GAMMA != 1 requires 0 < ALPHA < 1')
        if (np.abs(np.angle(np.repeat(z, np.abs(z) > eps))) <= alpha*np.pi).any():
            raise ValueError('|Arg(z)| <= alpha*pi')

    return np.vectorize(ml_, [np.float64])(z, alpha, beta, gama)


def ml_(z, alpha, beta, gama):
    # Target precision
    log_epsilon = np.log(1.e-15)
    # Inversion of the LT
    if np.abs(z) < 1.e-15:
        return 1/gamma(beta)
    else:
        return LTInversion(1, z, alpha, beta, gama, log_epsilon)



import numpy as np

import numpy.typing as npt

from scipy.special import gamma

from .model import Model
from .utils.mittag_leffler import ml as ml_accurate


def ml_fast(z: npt.NDArray[np.float64], a: float, b: float, inf: int = 100) -> npt.NDArray[np.float64]:
    z = np.asarray(z)
    result = np.zeros_like(z)
    for n in range(inf):
        term = (z ** n) / gamma(a * n + b)
        result += term
    return result


def ml_vec(z: npt.NDArray[np.float64], alpha: float, beta: float, inf: int = 100):
    k = np.arange(0, inf, dtype=int)[:, np.newaxis]
    return (z**k/gamma(alpha*k + beta)).sum(axis=0)


class FractionalSLS(Model):
    mode: str
    A: float
    E1: float
    T1: float
    CA: float
    B: float
    E2: float
    T2: float
    CB: float
    EK: float

    E: float
    T: float

    def __init__(self, config: dict[str, float], mode='fast') -> None:
        super().__init__()
        self.mode = mode
        if all(k in config for k in ['A', 'CA', 'B', 'CB', 'EK']):
            self._initialize_direct(config)
        elif all(k in config for k in ['A', 'E1', 'T1', 'E2', 'T2', 'B', 'EK']):
            self._initialize_computed(config)
        else:
            raise ValueError(
                "Invalid configuration: Provide either (A, CA, B, CB, EK) or (A, E1, T1, B, E2, T2, EK)")

    def _initialize_direct(self, config: dict[str, float]) -> None:
        required_attrs = ['A', 'CA', 'B', 'CB', 'EK']
        for attr in required_attrs:
            setattr(self, attr, config[attr])

    def _initialize_computed(self, config: dict[str, float]) -> None:
        required_attrs = ['A', 'E1', 'T1', 'E2', 'T2', 'B', 'EK']
        for attr in required_attrs:
            setattr(self, attr, config[attr])

    def make(self) -> None:
        if all(hasattr(self, attr) for attr in ['E1', 'T1', 'E2', 'T2']):
            self.CA = self.E1 * self.T1 ** self.A
            self.CB = self.E2 * self.T2 ** self.B

        self.T = (self.CA / self.CB) ** (1 / (self.A - self.B))
        self.E = self.CA / (self.T ** self.A)

    def run(self, dwell_size: int = 1001) -> None:
        self.make()
        self.time = self._get_time(self.D, self.L, dwell_size)
        u = np.heaviside
        self.strain = (self.I * self.time / self.D) * (1 - u(self.time - self.D, 0.5)) + \
            self.I * u(self.time - self.D, 0.5)
        modulus = np.zeros_like(self.time)

        if self.mode == 'fast':
            func = ml_fast
        elif self.mode == 'vectorized':
            func = ml_vec
        else:
            func = ml_accurate

        modulus[1:] = self.E * (self.time[1:] / self.T) ** (-self.B) * \
            func(-(self.time[1:] / self.T)
                 ** (self.A - self.B), self.A-self.B, 1-self.B) + self.EK

        self.stress = modulus * self.strain



import numpy as np
import numpy.typing as npt


class Model:
    I: float
    D: float
    L: float

    time: npt.NDArray[np.float64]
    strain: npt.NDArray[np.float64]
    stress: npt.NDArray[np.float64]

    def __init__(self):
        pass

    def __getitem__(self, key: str):
        try:
            return getattr(self, key)
        except AttributeError:
            raise KeyError(
                f"The attribute '{key}' does not exist in this model.")

    def _get_time(self, D: float, L: float, dwell_size: int) -> npt.NDArray[np.float64]:
        time = np.linspace(0, D, 1001, endpoint=False)
        time = np.hstack((time, D + np.linspace(0, L, dwell_size)))
        return time

    def set_test(self, config: dict[str, float]) -> None:
        if all(k in config for k in ['I', 'D', 'L']):
            self.I = config['I']
            self.D = config['D']
            self.L = config['L']
        else:
            raise ValueError("Provide I, D, and L for test config.")

    def make(self) -> None:
        pass

    def run(self, dwell_size: int = 1001) -> None:
        pass

    def trigger_force(self, trigger: float, dwell_size: int = 10001) -> None:
        while True:
            self.run(dwell_size=dwell_size)
            approach = self.get_approach()
            stress = approach["stress"]
            max_stress = stress[-1]
            if max_stress < 0:
                self.stress = -1 * self.time
                break
            if np.isclose(max_stress, trigger, rtol=1e-3):
                break
            if max_stress > trigger:
                rate = self.I / self.D
                trigger_idx = np.argmin(np.abs(stress - trigger))
                trigger_idx = max(trigger_idx, 1)
                self.D = self.time[trigger_idx]
                self.I = self.D * rate
            elif max_stress < trigger:
                scale = trigger / max_stress
                self.I *= scale
                self.D *= scale

    def get_approach(self) -> dict[str, npt.NDArray[np.float64]]:
        idx = np.argmin(np.abs(self.time - self.D))
        approach = {
            'time': self.time[:idx+1],
            'strain': self.strain[:idx+1],
            'stress': self.stress[:idx+1]
        }
        return approach

    def get_dwell(self) -> dict[str, npt.NDArray[np.float64]]:
        idx = np.argmin(np.abs(self.time - self.D))
        dwell = {
            'time': self.time[idx:],
            'strain': self.strain[idx:],
            'stress': self.stress[idx:]
        }
        return dwell



import numpy as np

from .model import Model


class SLS(Model):
    E1: float
    T1: float
    EK: float

    def __init__(self, config: dict[str, float]) -> None:
        super().__init__()
        required_attrs = [attr for attr in self.__annotations__]

        for attr in required_attrs:
            if attr in config:
                setattr(self, attr, config[attr])
            else:
                raise ValueError(
                    f"Missing required configuration for '{attr}'")

    def run(self, dwell_size: int = 1001) -> None:
        self.time = self._get_time(self.D, self.L, dwell_size)
        u = np.heaviside
        self.strain = (self.I * self.time / self.D) * (1 - u(self.time - self.D, 0.5)) + \
            self.I * u(self.time - self.D, 0.5)
        self.modulus = np.zeros_like(self.time)

        self.stress = self.I*(-self.E1*self.T1*(-np.exp(self.D/self.T1) + np.exp(self.time/self.T1))*u(-self.D + self.time, 0.5) + self.E1*self.time*np.exp(self.time/self.T1) - self.E1*(-self.D + self.T1*np.exp((self.D - self.time)/self.T1) - self.T1 + self.time)*np.exp(self.time/self.T1)
                              * u(-self.D + self.time, 0.5) - self.EK*self.T1*(-np.exp(self.D/self.T1) + np.exp(self.time/self.T1))*u(-self.D + self.time, 0.5) + self.EK*self.T1*np.exp(self.time/self.T1) - self.EK*self.T1)*np.exp(-self.time/self.T1)/self.D



import copy
from ..model import Model


def sweep(model: Model, x_to_y: tuple[str, str], values: list[float]) -> dict[float, Model]:
    sweep_dict = dict()
    x, y = x_to_y
    y_value = getattr(model, y)
    for val in values:
        new_model = copy.deepcopy(model)
        x_value = val * y_value
        setattr(new_model, x, x_value)
        new_model.make()
        new_model.run()
        sweep_dict[val] = new_model

    return sweep_dict


map_dict = {
    'A': '\\alpha',
    'B': '\\beta',
    'E1': 'E_1',
    'E2': 'E_2',
    'T1': '\\tau_1',
    'T2': '\\tau_2',
    'EK': 'E_K',
    'CA': 'c_{\\alpha}',
    'CB': 'c_{\\beta}',
    'I': '\\epsilon',
    'D': 't_{ind}',
    'L': 't_{dwell}',
}
